/*
 * E.S.O. - VLT project 
 * $Id: tRtd.C,v 1.2 2003/01/17 11:10:02 brighton Exp $
 *
 * tRtd.C - test RTD real-time updates by sending image and rapid frame
 *          
 * 
 * who             when       what
 * --------------  --------   ----------------------------------------
 * Allan Brighton  16 Jul 96  Created
 * pbiereic        16 Sept 97 Use semaphores for image event synchronization
 * pbiereic        01/03/01   More generalized version
 */

/************************************************************************
 *
 *   DESCRIPTION
 *     tRtd is used as a "test camera" for testing RTD image events.
 *     The options for tRtd can be set either via command line options
 *     or the RTD widget (tRtd.tcl) which is shown when RTD is started with the
 *     "-debug 1" option.
 *     tRtd generates image events at a defined speed (option -t). By default,
 *     it uses semaphore locking to avoid "image jitters" and other (nasty)
 *     side effects which can be seen when the option "-l 0" is set.
 *     tRtd generates the images in a sort of "movie" style, i.e. they continuously
 *     change so that the image update can be seen in the RTD image. A reference
 *     point at the position REF_PIXEL is generated which is used for analysis.
 *     Another point moves continuously across the image.
 *     
 *     Three different images can be generated with tRtd:
 *       o a continuously changing image pattern   
 *       o a star image (ngc1275.fits) with a moving area at the star position
 *         (XS, YS). This image is used for tests with "pick object"
 *       o a rapid frame. The rapid frame is displayed at a higher frequency
 *         than the main image (option "-t" value / RAPIDS). 
 *     
 *
 *   CAUTIONS
 *     tRtd uses image data of type "short". Other data types are currently not
 *     supported.
 */

#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define NTOHS(x) (ntohs(x))

#define LOCK_DELAY 5	// delay when all shm buffers are locked (in msec)
#define RAPIDS 5	// number of rapid frames displayed before main image
#define REF_PIXEL 20	// center of reference pixel
#define REF_SIZE 3	// size of reference area (MUST be an odd number)
#define COUNT_INC 3	// count increment after an image was displayed
#define MAX_NX 4000   	// max #of pixels in x
#define MAX_NY 4000  	// max #of pixels in x

// defines if FITS file "ngc1275.fits" is used (option -I).
// ngc1275.fits is traditionally available on rtdimg/images with the
// parameters shown in the defines below.
#define DFILE		"ngc1275.fits"
#define OFFSET		11520   // byte offset to where the image data start
#define DATA_TYPE	short	// data type
#define NUMBYTE		sizeof(DATA_TYPE) // number of bytes in DATA_TYPE
#define WINSIZE		50      // size of the moving square around the star
#define XSIZE		353     // width of the image
#define YSIZE		353     // height of the image
// start point of the rectangular which contains the star
#define XS		277 - WINSIZE / 2
#define YS		196 - WINSIZE / 2
// max. shift of the rectangular (+/- MAXSHIFT)
#define MAXSHIFT	3

#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "rtdSem.h"
#include "error.h"
#include "rtdImageEvent.h"

extern char *optarg;

// structures for multibuffering/semaphore convenience routines
static rtdShm shmMain;
static rtdShm shmRapid;

static DATA_TYPE data[MAX_NX * MAX_NY];
static DATA_TYPE lineBuffer[MAX_NX * 3];

/*
 * cleanup and exit
 */
static void cleanup(int i=0)
{
    i = 1;  // to keep the compiler quite
    rtdShmDelete(&shmMain);
    rtdShmDelete(&shmRapid);
    exit(1);
}

/*
 * print error message and exit
 */
static void errexit(const char* msg1, const char* msg2 = "")
{
    printf("%s %s\n", msg1, msg2);
    cleanup(1);
}

/*
 * print usage and exit
 */
void usage()
{
    char* usage = "usage: tRtd \n"
	"\t -v Verbose flag\n"
	"\t -c Camera name\n"
	"\t -t Update interval in msecs\n"
	"\t -W Main frame width\n"
	"\t -H Main frame height\n"
	"\t -w Rapid frame width\n"
	"\t -h Rapid frame height\n"
	"\t -x Rapid frame start x\n"
	"\t -y Rapid frame start y\n"
	"\t -f Rapid frame Id\n"
	"\t -l Use semaphore locking\n"
	"\t -b Number of shared memory buffers\n"
	"\t -I Directory for file ngc1275.fits, used for generating the image\n";
    printf("%s\n", usage);
    exit(1);
}

/*
 * generate a reference pixel
 */
void gen_ref(DATA_TYPE* p, int width, int height, int ref)
{
    if ((REF_SIZE % 2) == 0)
	return;

    int ref_size = REF_SIZE / 2;
    DATA_TYPE valmin = -10000;
    DATA_TYPE valmax =  10000;

    // generate a n x n area around the reference pixel
    if (ref <= ref_size || width < ref + ref_size || height < ref + ref_size)
	return;

    ref--;  // for indexing
    for (int k = ref - ref_size; k <= ref + ref_size; k++)
	for (int i = ref - ref_size; i <= ref + ref_size; i ++)
	    *(p + k * width + i) = NTOHS(valmax);
  
    *(p + ref * width + ref) = NTOHS(valmin);
}

/*
 * generate some dummy image data based on the given count.
 * XXX needs to be "static", otherwise it hangs after a while on Linux
 */
static void gen_image(DATA_TYPE* p, int width, int height, int count)
{
    for (int k = 0; k < height; k++) {
	int cnt = (count + COUNT_INC) % width;
	memcpy(p + k * width, lineBuffer + cnt, width * NUMBYTE);
    }
    // we need some reference pixels for tests
    gen_ref(p, width, height, REF_PIXEL);
    gen_ref(p, width, height, count);
    gen_ref(p, width, height, width - count);
}

/*
 * print image event info
 */    
void printImageInfo(char *text, int verbose, int numShm, rtdIMAGE_INFO *imageInfo, int count)
{
    if (! verbose)
	return;

    printf("update #%d %s frame: frameId=%d, %dx%d+%d+%d, semId=%d, buffer %d(%d)\n", 
	   count, text, imageInfo->frameId,
	   imageInfo->xPixels, imageInfo->yPixels, imageInfo->frameX, imageInfo->frameY,
	   imageInfo->semId, imageInfo->shmNum + 1, numShm);
}

/*
 * fill image event info
 */    
void fillImageInfo(
    rtdIMAGE_INFO *imageInfo,
    int width,
    int height,
    int count, 
    int type,
    int x,
    int y,
    int frameId)
{
    // clear info in rtdIMAGE_INFO
    memset(imageInfo, '\0', sizeof(rtdIMAGE_INFO));

    imageInfo->dataType = type;
    imageInfo->frameX   = x;
    imageInfo->frameY   = y;
    imageInfo->xPixels  = width;
    imageInfo->yPixels  = height;
    imageInfo->frameId  = frameId;

  // add some dummy world coordinates for testing
    imageInfo->ra       = 49.951;
    imageInfo->dec      =  41.5117;
    imageInfo->secpix   = 1.70127;
    imageInfo->xrefpix  = width/2.0;
    imageInfo->yrefpix  = height/2.0 ;
    imageInfo->rotate   = double(count % 360);
    imageInfo->equinox  = 2000;
    imageInfo->epoch    = 1957.97;
    strcpy(imageInfo->proj, "-TAN");

#ifdef DEBUG
    // add some dummy detector information
    if (frameId == 0) {
	imageInfo->startX    = count;
	imageInfo->startY    = count;
	imageInfo->binningX  = 2;
	imageInfo->binningY  = 2;
	imageInfo->lowCut    = 0;
	imageInfo->highCut   = 100;
    }
#endif
}

void copyArrayToImage(DATA_TYPE *array, DATA_TYPE *image, int xs, int ys)
{
    DATA_TYPE *s1, *s2;
    for (int idx = 0; idx < WINSIZE; idx++) {
	s1 = array + idx * WINSIZE;
	s2 = image + (ys - 1 + idx) * XSIZE + xs - 1;
	memcpy(s2, s1, WINSIZE * NUMBYTE);
    }
}

void copyImageToArray(DATA_TYPE *image, DATA_TYPE *array, int xs, int ys)
{
    DATA_TYPE *s1, *s2;
    for (int idx = 0; idx < WINSIZE; idx++) {
	s1 = array + idx * WINSIZE;
	s2 = image + (ys - 1 + idx) * XSIZE + xs - 1;
	memcpy(s1, s2, WINSIZE * NUMBYTE);
    }
}

/*
 * get a random shift with a size of -MAXSHIFT to +MAXSHIFT
 */
int getRandomShift()
{
    int random = rand();
    double shift, r;
    r = (double)(RAND_MAX / 2.);
    shift = ((random - r) / r) * MAXSHIFT;
    return (int)shift;
}

/* 
 * Main:
 */
main(int argc, char** argv) 
{
    rtdIMAGE_EVT_HNDL	eventHndl;		// image event handle
    rtdIMAGE_INFO		imageInfo;		// image info structure

    char  *rtd_camera = 	getenv("RTD_CAMERA");	// default name of camera
    int	verbose =	0;	// verbose flag: if true, print messages
    int	dataType =	sizeof(DATA_TYPE) * 8;	// default data type
    int	delay =		500;	// default delay between updates
    int	numShm =	2;	// default number of shm buffers to use

    int	main_id  =	0;	// default frame id
    int	rapid_id =	0;	// default frame id
    int	lock =		1;	// Flag: use semaphore locking
    
    int	rapid_x =	0;	// default rapid frame start x
    int	rapid_y =	0;	// default rapid frame start y
    int	main_width =	255;	// default image width
    int	main_height =	255;	// default image height
    int	rapid_width =	255;	// default image width
    int	rapid_height =	255;	// default image height

    int 	countMain  = 	COUNT_INC+1;	// counter
    int	countRapid = 	COUNT_INC+1;	// counter
    int 	countM  = 	0;	// counter
    int	countR = 	0;	// counter
    int 	countMSign  = 	1;	// counter sign
    int	countRSign = 	1;	// counter sign
    int 	indexMain = 	0;	// index for shared memory
    int 	indexRapid = 	0;	// index for shared memory
    int	status =	0;	// status
    char  errBuf[256];

    int	useFits =	0;	// use Fits file for generating the image
    char	*fitsDir =	NULL;	// Directory where we can find ngc1275.fits
    int	arraySize;
    DATA_TYPE	*array;
    DATA_TYPE	*array2;

    int c;
    while ((c = getopt(argc, argv, "x:y:w:h:W:H:f:v:c:t:b:l:I:")) != -1) {
	switch(c) {
	case 'x': 
	    rapid_x = 	atoi(optarg); break;
	case 'y': 
	    rapid_y = 	atoi(optarg); break;
	case 'W': 
	    main_width = 	atoi(optarg); break;
	case 'H': 
	    main_height = 	atoi(optarg); break;
	case 'w': 
	    rapid_width = 	atoi(optarg); break;
	case 'h': 
	    rapid_height = 	atoi(optarg); break;
	case 'f': 
	    rapid_id = 	atoi(optarg); break;
	case 'v': 
	    verbose = 	atoi(optarg); break;
	case 'c': 
	    rtd_camera = 	optarg;  break;
	case 't': 
	    delay = 	atoi(optarg); break;
	case 'b': 
	    numShm = 	atoi(optarg); break;
	case 'l': 
	    lock = 		atoi(optarg); break;
	case 'I': 
	    fitsDir = 	optarg; useFits = 1; break;
	default: 
	    usage();
	}
    }

    // print arguments
    if (verbose) {
	for (int i = 0; i < argc; i++)
	    printf("%s ", argv[i]);
	printf("\n");
    }

    if (! useFits) {
	// check arguments
	if (rapid_x < 0 || rapid_y < 0 || rapid_width < 0 || rapid_height < 0 || 
	    main_width <= 0 || main_height <= 0 || rapid_id < 0 || delay < 0 || 
	    numShm  <= 0)
	    errexit("wrong argument");
	// limit ranges
	main_width   = MIN(main_width, MAX_NX);
	main_height  = MIN(main_height, MAX_NY);
	rapid_width  = MIN(rapid_width, MAX_NX);
	rapid_height = MIN(rapid_height, MAX_NY);
	rapid_x      = MIN(rapid_x, MAX_NX);
	rapid_y      = MIN(rapid_y, MAX_NY);
    }
    else {
	main_width  = XSIZE;
	main_height = YSIZE;
	rapid_width = rapid_height = 10;  // dummy (rapids are not used)
	rapid_id = 0;
    }

    if (! rtd_camera) 
	errexit("please use '... -c camera' or 'setenv RTD_CAMERA ...' first");
    
    if (rtdInitImageEvt(rtd_camera, &eventHndl, errBuf) != 0) {
	errexit("Could not initialize image event: check if rtdServer is running");
    }

    // clean up shared memory on exit
    signal(SIGINT,  cleanup);
    signal(SIGTERM, cleanup);
    signal(SIGHUP,  cleanup);

    // generate a line with color values
    DATA_TYPE *ptmp = lineBuffer;
    DATA_TYPE v = -main_width / 2;
    for (int i = 0; i < MAX_NX * 3; i++, v++)
	*ptmp++ = NTOHS(v);
  
    if (useFits) {
	char buf[4096];
	sprintf(buf, "%s/%s", fitsDir, DFILE);
	if (verbose)
	    printf("Using FITS file: %s\n", buf);

	// get data from FITS file
	int fd = open(buf, O_RDONLY);
	if (fd < 0)
	    errexit("error opening file", buf);

	if ((read(fd, data, OFFSET)) < 0)
	    errexit("error reading file", buf);

	if ((read(fd, data, XSIZE * YSIZE * NUMBYTE)) < 0)
	    errexit("error reading file", buf);

	close(fd);

	gen_ref(data, main_width, main_height, REF_PIXEL);

	// allocate buffers for saving and copying data
	arraySize = WINSIZE * WINSIZE * NUMBYTE;
	array  = (DATA_TYPE *)malloc(arraySize);
	array2 = (DATA_TYPE *)malloc(arraySize);

	// save the rectangular around the star into array
	copyImageToArray(data, array, XS, YS);
    }
    int xs = XS, ys = YS;

    /*
     * create shared memory areas
     */
    if (rtdShmCreate(numShm, &shmMain,  main_width,  main_height,  dataType) == -1 ||
	rtdShmCreate(numShm, &shmRapid, rapid_width, rapid_height, dataType) == -1)
	errexit("error creating shared memory");

    /*
     * 1. Generate the image.
     * 2. Wait until the next shared memory buffer is unlocked.
     * 3. Copy data to shared memory and lock it.
     * 4. Fill the image event info structure.
     * 5. Fill the image information fields with semaphore/shm info.
     * 6. Send the image information.
     * 7. Delay before next event.
     * 8. Cycle shm index.
     */

    while ( 1 ) {
	if ( ! useFits) {
	    gen_image(data, main_width, main_height, countMain);
	}
	else {
	    ys = YS + getRandomShift();
	    xs = XS + getRandomShift();
     
	    // copy array starting at xs,ys to the area starting at XS, YS
	    copyImageToArray(data, array2, xs, ys);
	    copyArrayToImage(array2, data, XS, YS);
	}

	while (rtdShmFill(indexMain, (char *)data, &shmMain, 0) == -1)
	    rtdSleep((int)LOCK_DELAY);

	int cnt = countMain + countMSign * COUNT_INC;
	if (cnt >= main_width || cnt <= COUNT_INC)
	    countMSign = -countMSign;
	countMain = countMain + countMSign * COUNT_INC;

	fillImageInfo(&imageInfo, main_width, main_height,
		      countMain, dataType, 0, 0, 0);

	if (rtdShmStruct(indexMain, &imageInfo, &shmMain) == -1)
	    errexit("rtdShmStruct failed");

	if ( ! lock)
	    rtdSemReset(imageInfo.semId, indexMain);

	printImageInfo("Main", verbose, numShm, &imageInfo, ++countM);
	if (rtdSendImageInfo(&eventHndl, &imageInfo, errBuf) == RTD_ERROR)
	    errexit("rtdSendImageInfo failed");

	if (useFits) {
	    // restore the area starting at XS, YS
	    copyArrayToImage(array, data, XS, YS);
	}

	indexMain = (indexMain + 1) % numShm;
	rtdSleep(delay);


	if (rapid_id == 0)
	    continue;



	/*
	 * Rapid frame
	 */

	for (int i = 0; i < RAPIDS; i++) {
	    gen_image(data, rapid_width, rapid_height, countRapid);

	    while (rtdShmFill(indexRapid, (char *)data, &shmRapid, 0) == -1)
		rtdSleep((int)LOCK_DELAY);

	    countRapid = (countRapid + 1) % rapid_width;

	    int cnt = countRapid + countRSign * COUNT_INC;
	    if (cnt >= rapid_width || cnt <= COUNT_INC)
		countRSign = -countRSign;
	    countRapid = countRapid + countRSign * COUNT_INC;
      
	    fillImageInfo(&imageInfo, rapid_width, rapid_height,
			  countRapid, dataType, 0,0, rapid_id);
      
	    if (rtdShmStruct(indexRapid, &imageInfo, &shmRapid) == -1)
		errexit("rtdShmStruct failed");
      
	    printImageInfo("Rapid", verbose, numShm, &imageInfo, ++countR);
	    if (rtdSendImageInfo(&eventHndl, &imageInfo, errBuf) == RTD_ERROR)
		errexit("rtdSendImageInfo failed");
      
	    indexRapid = (indexRapid + 1) % numShm;
	    rtdSleep(delay / RAPIDS);
	}

    }  // -- end while( 1 )
    return 0;
}

